"""Checker for quote handling on string literals."""

from __future__ import annotations

import tokenize
from abc import abstractproperty
from typing import ClassVar, FrozenSet, Iterator, List, Optional, Sequence, Set, TYPE_CHECKING, Tuple, Type

from typing_extensions import Protocol

if (TYPE_CHECKING):
	import ast


try:
	try:
		from importlib.metadata import version
	except ModuleNotFoundError:  # python < 3.8 use polyfill
		from importlib_metadata import version  # type: ignore
	package_version = version(__package__)
except Exception:
	package_version = 'unknown'


IGNORE_TOKENS = frozenset((
	tokenize.ENCODING,
	tokenize.NEWLINE,
	tokenize.INDENT,
	tokenize.DEDENT,
	tokenize.NL,
	tokenize.COMMENT,
))

OPEN_BRACKET = frozenset(('(', '[', '{'))
CLOSE_BRACKET = frozenset((')', ']', '}'))


class Options(Protocol):
	"""Protocol for options."""

	literal_include_name: bool


class Message(Protocol):
	"""Messages."""

	@abstractproperty
	def code(self) -> str:
		...

	def text(self, **kwargs) -> str:
		...


LogicalResult = Tuple[Tuple[int, int], str]  # (line, column), text
PhysicalResult = Tuple[int, str]  # (column, text)
ASTResult = Tuple[int, int, str, Type]  # (line, column, text, Type)


class Checker:
	"""Base class for checkers."""

	name: ClassVar[str] = __package__.replace('_', '-')
	version: ClassVar[str] = package_version
	plugin_name: ClassVar[str]

	@classmethod
	def parse_options(cls, options: Options) -> None:
		cls.plugin_name = (' (' + cls.name + ')') if (options.literal_include_name) else ''

	def _logical_token_message(self, token: tokenize.TokenInfo, message: Message, **kwargs) -> LogicalResult:
		return (token.start, f'{message.code}{self.plugin_name} {message.text(**kwargs)}')

	def _pyhsical_token_message(self, token: tokenize.TokenInfo, message: Message, **kwargs) -> PhysicalResult:
		return (token.start[1], f'{message.code}{self.plugin_name} {message.text(**kwargs)}')

	def _ast_token_message(self, token: tokenize.TokenInfo, message: Message, **kwargs) -> ASTResult:
		return (token.start[0], token.start[1], f'{message.code}{self.plugin_name} {message.text(**kwargs)}', type(self))

	def _ast_node_message(self, node: ast.AST, message: Message, **kwargs) -> ASTResult:
		return (node.lineno, node.col_offset, f'{message.code}{self.plugin_name} {message.text(**kwargs)}', type(self))


class LiteralChecker(Checker):
	"""Base class for literal checkers."""

	tokens: Sequence[tokenize.TokenInfo]
	_docstring_tokens: Optional[FrozenSet[tokenize.TokenInfo]]

	def __init__(self, logical_line: str, tokens: Sequence[tokenize.TokenInfo]) -> None:
		self.tokens = tokens
		self._docstring_tokens = None

	@property
	def docstring_tokens(self) -> FrozenSet[tokenize.TokenInfo]:
		"""Find docstring tokens, which are initial strings or strings immediately after class or function defs."""
		if (self._docstring_tokens is None):
			docstrings: Set[tokenize.TokenInfo] = set()
			expect_docstring = True
			expect_colon = False
			bracket_depth = 0
			for token in self.tokens:
				if (token.type in IGNORE_TOKENS):
					continue
				if (tokenize.STRING == token.type):
					if (expect_docstring):
						docstrings.add(token)
				else:
					expect_docstring = False
					if ((tokenize.NAME == token.type) and (token.string in ('class', 'def'))):
						expect_colon = True
						bracket_depth = 0
					elif (tokenize.OP == token.type):
						if (':' == token.string):
							if (0 == bracket_depth):
								if (expect_colon):
									expect_docstring = True
								expect_colon = False
						elif (token.string in OPEN_BRACKET):
							bracket_depth += 1
						elif (token.string in CLOSE_BRACKET):
							bracket_depth -= 1
			self._docstring_tokens = frozenset(docstrings)
		return self._docstring_tokens

	def _process_literals(self, tokens: Sequence[tokenize.TokenInfo]) -> Iterator[LogicalResult]:
		raise NotImplementedError()

	def __iter__(self) -> Iterator[LogicalResult]:
		"""Primary call from flake8, yield error messages."""
		continuation: List[tokenize.TokenInfo] = []
		for token in self.tokens:
			if (token.type in IGNORE_TOKENS):
				continue

			if (tokenize.STRING == token.type):
				continuation.append(token)
				continue

			for message in self._process_literals(continuation):
				yield message
			continuation = []

		for message in self._process_literals(continuation):
			yield message



"""Checker for quote handling on string literals."""

from __future__ import annotations

import enum
from typing import ClassVar, Iterator, NamedTuple, Optional, Sequence, TYPE_CHECKING, Tuple

import flake8_literal

from . import checker

if (TYPE_CHECKING):
	import tokenize
	from flake8.options.manager import OptionManager


class Message(enum.Enum):
	"""Messages."""

	USE_SINGLE = (1, 'Use single quotes for string')
	USE_DOUBLE = (2, 'Use double quotes for string')
	MULTILINE_USE_SINGLE = (3, 'Use single quotes for multiline string')
	MULTILINE_USE_DOUBLE = (4, 'Use double quotes for multiline string')
	DOCSTRING_USE_SINGLE = (5, 'Use single quotes for docstring')
	DOCSTRING_USE_DOUBLE = (6, 'Use double quotes for docstring')
	DOCSTRING_USE_TRIPLE_SINGLE = (7, 'Use triple single quotes for docstring')
	DOCSTRING_USE_TRIPLE_DOUBLE = (8, 'Use triple double quotes for docstring')
	SWITCH_TO_DOUBLE = (11, 'Use double quotes for string to avoid escaped single quote')
	SWITCH_TO_SINGLE = (12, 'Use single quotes for string to avoid escaped double quote')
	UNNECESSARY_ESCAPE_SINGLE = (13, 'Escaped single quote is not necessary')
	UNNECESSARY_ESCAPE_DOUBLE = (14, 'Escaped double quote is not necessary')
	MATCH_CONTINUATION_DOUBLE = (15, 'Use double quotes for continuation strings to match')
	MATCH_CONTINUATION_SINGLE = (16, 'Use single quotes for continuation strings to match')
	UNNECESSARY_RAW = (21, 'Remove raw prefix when not using escapes')
	USE_RAW_PREFIX = (22, 'Use raw prefix to avoid escaped slash')

	@property
	def code(self) -> str:
		return (flake8_literal.quote_checker_prefix + str(self.value[0]).rjust(6 - len(flake8_literal.quote_checker_prefix), '0'))

	def text(self, **kwargs) -> str:
		return self.value[1].format(**kwargs)


class QuoteType(enum.Enum):
	"""Quote type enum."""

	SINGLE = 'single'
	DOUBLE = 'double'

	@classmethod
	def from_str(cls, value: str) -> Optional[QuoteType]:
		for member in cls.__members__.values():
			if (value.lower() == member.value):
				return member
		return None


QUOTE = {
	QuoteType.SINGLE: "'",
	QuoteType.DOUBLE: '"',
}

OTHER_QUOTE = {
	QuoteType.SINGLE: '"',
	QuoteType.DOUBLE: "'",
}

USE_QUOTE_MESSAGE = {
	QuoteType.SINGLE: Message.USE_SINGLE,
	QuoteType.DOUBLE: Message.USE_DOUBLE,
}

AVOID_ESCAPE_MESSAGE = {
	QuoteType.SINGLE: Message.SWITCH_TO_DOUBLE,
	QuoteType.DOUBLE: Message.SWITCH_TO_SINGLE,
}

UNNECESSARY_ESCAPE_MESSAGE = {
	QuoteType.SINGLE: Message.UNNECESSARY_ESCAPE_DOUBLE,
	QuoteType.DOUBLE: Message.UNNECESSARY_ESCAPE_SINGLE,
}

UNNECESSARY_OTHER_ESCAPE_MESSAGE = {
	QuoteType.SINGLE: Message.UNNECESSARY_ESCAPE_SINGLE,
	QuoteType.DOUBLE: Message.UNNECESSARY_ESCAPE_DOUBLE,
}

MULTILINE_USE_QUOTE_MESSAGE = {
	QuoteType.SINGLE: Message.MULTILINE_USE_SINGLE,
	QuoteType.DOUBLE: Message.MULTILINE_USE_DOUBLE,
}

DOCSTRING_USE_QUOTE_MESSAGE = {
	QuoteType.SINGLE: Message.DOCSTRING_USE_SINGLE,
	QuoteType.DOUBLE: Message.DOCSTRING_USE_DOUBLE,
}

DOCSTRING_USE_TRIPLE_MESSAGE = {
	QuoteType.SINGLE: Message.DOCSTRING_USE_TRIPLE_SINGLE,
	QuoteType.DOUBLE: Message.DOCSTRING_USE_TRIPLE_DOUBLE,
}

MATCH_CONTINUATION_MESSAGE = {
	QuoteType.SINGLE: Message.MATCH_CONTINUATION_DOUBLE,
	QuoteType.DOUBLE: Message.MATCH_CONTINUATION_SINGLE,
}


class Options(checker.Options):
	"""Protocol for options."""

	literal_inline: str
	literal_multiline: str
	literal_docstring: str
	literal_avoid_escape: bool


class Config(NamedTuple):
	"""Config options."""

	inline: QuoteType
	multiline: QuoteType
	docstring: QuoteType
	avoid_escape: bool


class QuoteChecker(checker.LiteralChecker):
	"""Check string literals for proper quotes."""

	config: ClassVar[Config]

	@classmethod
	def add_options(cls, option_manager: OptionManager) -> None:
		option_manager.add_option('--literal-inline-quotes', default='single',
		                          action='store', parse_from_config=True,
		                          choices=('single', 'double'), dest='literal_inline',
		                          help='Quote to use for inline string literals (default: single)')
		option_manager.add_option('--literal-multiline-quotes', default='single',
		                          action='store', parse_from_config=True,
		                          choices=('single', 'double'), dest='literal_multiline',
		                          help='Quote to use for multiline string literals (default: single)')
		option_manager.add_option('--literal-docstring-quotes', default='double',
		                          action='store', parse_from_config=True,
		                          choices=('single', 'double'), dest='literal_docstring',
		                          help='Quote to use for docstrings (default: double)')
		option_manager.add_option('--literal-avoid-escape', default=True, action='store_true',
		                          parse_from_config=True,
		                          help='Avoid escapes in inline string literals when possible (enabled by default)')
		option_manager.add_option('--literal-no-avoid-escape', default=None, action='store_false',
		                          parse_from_config=False, dest='literal_avoid_escape',
		                          help='Disable escape avoidance in inline string literals')
		option_manager.add_option('--literal-include-name', default=False, action='store_true',
		                          parse_from_config=True, dest='literal_include_name',
		                          help='Include plugin name in messages (enabled by default)')
		option_manager.add_option('--literal-no-include-name', default=None, action='store_false',
		                          parse_from_config=False, dest='literal_include_name',
		                          help='Remove plugin name from messages')

	@classmethod
	def parse_options(cls, options: Options) -> None:  # type: ignore[override]
		super().parse_options(options)
		cls.config = Config(inline=QuoteType.from_str(options.literal_inline) or QuoteType.SINGLE,
		                    multiline=QuoteType.from_str(options.literal_multiline) or QuoteType.SINGLE,
		                    docstring=QuoteType.from_str(options.literal_docstring) or QuoteType.DOUBLE,
		                    avoid_escape=options.literal_avoid_escape)

	def _process_literals(self, tokens: Sequence[tokenize.TokenInfo]) -> Iterator[Tuple[Tuple[int, int], str]]:
		if (not tokens):
			return

		desired = QUOTE[self.config.inline]
		other = OTHER_QUOTE[self.config.inline]
		needs_other = False
		cant_switch = False
		should_switch = set()
		others = []
		for token in tokens:
			quote = token.string[-1]
			prefix = token.string[:token.string.index(quote)].lower()
			string = token.string[len(prefix):]

			if (token in self.docstring_tokens):  # docstring
				if ((quote == QUOTE[self.config.docstring]) and (string[0:3] == (quote * 3))):
					continue
				if (quote == QUOTE[self.config.docstring]):
					yield self._logical_token_message(token, DOCSTRING_USE_TRIPLE_MESSAGE[self.config.docstring])
				else:
					yield self._logical_token_message(token, DOCSTRING_USE_QUOTE_MESSAGE[self.config.docstring])

			elif (string[0:3] == (quote * 3)):  # multiline
				if (quote == QUOTE[self.config.multiline]):
					continue
				yield self._logical_token_message(token, MULTILINE_USE_QUOTE_MESSAGE[self.config.multiline])

			else:  # inline
				contents = string[1:-1]

				if (quote == desired):  # check for escapes
					if ((not self.config.avoid_escape) or ('r' in prefix)):
						continue
					if (other in contents):
						cant_switch = True
					if ((desired in contents) and (other in contents)):
						continue  # both quotes used, nothing to do
					if (desired in contents):
						needs_other = True
						should_switch.add(token)
						yield self._logical_token_message(token, AVOID_ESCAPE_MESSAGE[self.config.inline])
					if ((other in contents) and (('\\' + other) in contents.replace(r'\\', ''))):
						yield self._logical_token_message(token, UNNECESSARY_ESCAPE_MESSAGE[self.config.inline])
				else:
					if ((desired in contents) and (other not in contents) and self.config.avoid_escape):
						if (('\\' + desired) in contents.replace(r'\\', '')):
							yield self._logical_token_message(token, UNNECESSARY_OTHER_ESCAPE_MESSAGE[self.config.inline])
						needs_other = True
						continue
					others.append(token)
		if ((1 < len(tokens)) and needs_other and (not cant_switch)):
			for token in tokens:
				quote = token.string[-1]
				if ((quote != desired) or (token.string[-3:] == (quote * 3)) or (token in should_switch)):
					continue
				yield self._logical_token_message(token, MATCH_CONTINUATION_MESSAGE[self.config.inline])
		else:
			for token in others:
				yield self._logical_token_message(token, USE_QUOTE_MESSAGE[self.config.inline])



"""\
flake8-string package.

Checks for literal string usage: quotes, format, f-strings
"""

plugin_prefix = 'LIT'

quote_checker_prefix = f'{plugin_prefix}0'
raw_checker_prefix = f'{plugin_prefix}1'



"""Checker for raw string literals."""

from __future__ import annotations

import ast
import enum
import tokenize
from typing import ClassVar, Iterator, List, NamedTuple, Optional, Sequence, Set, TYPE_CHECKING, Tuple

import flake8_literal

from . import checker

if (TYPE_CHECKING):
	from flake8.options.manager import OptionManager


class Message(enum.Enum):
	"""Messages."""

	UNNECESSARY_RAW = (1, 'Remove raw prefix when not using escapes')
	USE_RAW_FOR_SLASH = (2, 'Use raw prefix to avoid escaped slash')
	USE_RAW_FOR_REGEX = (3, 'Use raw prefix for re pattern')

	@property
	def code(self) -> str:
		return (flake8_literal.raw_checker_prefix + str(self.value[0]).rjust(6 - len(flake8_literal.raw_checker_prefix), '0'))

	def text(self, **kwargs) -> str:
		return self.value[1].format(**kwargs)


class RePatternRaw(enum.Enum):
	"""Raw regex type option enum."""

	AVOID = 'avoid'
	ALLOW = 'allow'
	ALWAYS = 'always'

	@classmethod
	def from_str(cls, value: str) -> Optional[RePatternRaw]:
		for member in cls.__members__.values():
			if (value.lower() == member.value):
				return member
		return None


class Options(checker.Options):
	"""Protocol for options."""

	literal_re_pattern_raw: str
	literal_avoid_escape: bool


class Config(NamedTuple):
	"""Config options."""

	re_pattern_raw: RePatternRaw
	avoid_escape: bool


RE_METHODS = frozenset((
	'compile',
	'search',
	'match',
	'fullmatch',
	'split',
	'findall',
	'finditer',
	'sub',
	'subn',
	'escape',
))


_ast_string_classes: List[type] = [ast.Constant]
try:
	_ast_string_classes.append(ast.Str)  # ast.Str deprecated in 3.8
	_ast_string_classes.append(ast.Bytes)  # ast.Bytes deprecated in 3.8
except AttributeError:
	pass
AST_STRINGS = tuple(_ast_string_classes)


class RawChecker(checker.Checker):
	"""Check string literals for proper raw usage."""

	config: ClassVar[Config]

	tokens: Sequence[tokenize.TokenInfo]
	re_arguments: Set[Tuple[int, int]]

	@classmethod
	def add_options(cls, option_manager: OptionManager) -> None:
		option_manager.add_option('--literal-re-pattern-raw', default='allow',
		                          action='store', parse_from_config=True,
		                          choices=('avoid', 'allow', 'always'), dest='literal_re_pattern_raw',
		                          help='Use raw strings for regular expressions (default:allow)')

	@classmethod
	def parse_options(cls, options: Options) -> None:  # type: ignore[override]
		super().parse_options(options)
		cls.config = Config(re_pattern_raw=RePatternRaw.from_str(options.literal_re_pattern_raw) or RePatternRaw.ALLOW,
		                    avoid_escape=options.literal_avoid_escape)

	def __init__(self, tree: ast.AST, file_tokens: Sequence[tokenize.TokenInfo]) -> None:
		super().__init__()
		self.tokens = file_tokens
		self.re_arguments = self._find_re_arguments(tree)

	def _find_re_arguments(self, tree: ast.AST) -> Set[Tuple[int, int]]:
		re_arguments: Set[Tuple[int, int]] = set()

		def _add_re_argument(node: ast.AST) -> None:
			if (isinstance(node, AST_STRINGS)):
				re_arguments.add((node.lineno, node.col_offset))
			if (isinstance(node, (ast.BinOp, ast.JoinedStr))):
				for child in ast.iter_child_nodes(node):
					_add_re_argument(child)
			if (isinstance(node, ast.Call) and isinstance(node.func, ast.Attribute)):
				_add_re_argument(node.func.value)

		def _process_node(node: ast.AST, modules: Set[str], functions: Set[str]) -> None:
			modules = set(modules)
			functions = set(functions)
			for child in ast.iter_child_nodes(node):
				if (isinstance(child, ast.Import)):
					for name in child.names:
						if ('re' == name.name):
							modules.add(name.asname if (name.asname is not None) else name.name)
				elif (isinstance(child, ast.ImportFrom)):
					if ('re' == child.module):
						for name in child.names:
							if (name.name in RE_METHODS):
								functions.add(name.asname if (name.asname is not None) else name.name)
				elif (isinstance(child, ast.Call)):
					function = child.func
					if (isinstance(function, ast.Name) and (function.id in functions)):  # direct function call
						if (child.args):
							_add_re_argument(child.args[0])
					elif ((isinstance(function, ast.Attribute)) and (function.attr in RE_METHODS)):	 # function is an attribute of something
						if (isinstance(function.value, ast.Name) and (function.value.id in modules)):  # function in re module
							if (child.args):
								_add_re_argument(child.args[0])
				_process_node(child, modules, functions)

		_process_node(tree, set(), set())
		return re_arguments

	def _process_literals(self, tokens: Sequence[tokenize.TokenInfo]) -> Iterator[checker.ASTResult]:
		if (not tokens):
			return
		re_pattern = (tokens[0].start in self.re_arguments)
		for token in tokens:
			quote = token.string[-1]
			prefix = token.string[:token.string.index(quote)].lower()
			string = token.string[len(prefix):]

			if (string[0:3] == (quote * 3)):  # multiline
				contents = string[3:-3]
			else:  # inline
				contents = string[1:-1]

			if ('r' in prefix):
				if ((not ('\\' in contents)) and ((not re_pattern) or (RePatternRaw.AVOID == self.config.re_pattern_raw))):
					yield self._ast_token_message(token, Message.UNNECESSARY_RAW)
			else:
				if ((r'\\' in contents) and ('\\' not in contents.replace(r'\\', '')) and self.config.avoid_escape):
					trail_count = 0
					test_contents = contents
					while test_contents.endswith(r'\\'):
						test_contents = test_contents[:-2]
						trail_count += 1
					if (0 == (trail_count % 2)):  # raw strings can't end in an odd number of backslashes
						yield self._ast_token_message(token, Message.USE_RAW_FOR_SLASH)
						continue
				if (re_pattern and (RePatternRaw.ALWAYS == self.config.re_pattern_raw)):
					yield self._ast_token_message(token, Message.USE_RAW_FOR_REGEX)

	def __iter__(self) -> Iterator[checker.ASTResult]:
		"""Primary call from flake8, yield error messages."""
		continuation: List[tokenize.TokenInfo] = []
		for token in self.tokens:
			if (token.type in checker.IGNORE_TOKENS):
				continue

			if (tokenize.STRING == token.type):
				continuation.append(token)
				continue

			for message in self._process_literals(continuation):
				yield message
			continuation = []

		for message in self._process_literals(continuation):
			yield message



"""Define PyPI package."""

import flake8_literal

import setuptools


with open('README.md', 'r') as readme_file:
	long_description = readme_file.read()

setuptools.setup(
	name='flake8-literal',
	version='1.2.6',
	author='Peter Linss',
	author_email='pypi@linss.com',
	description='Flake8 string literal validation',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/plinss/flake8-literal/',

	packages=['flake8_literal'],
	package_data={'flake8_literal': ['py.typed']},

	install_requires=[
		'flake8>=3.8.0,<6.0',
		'importlib_metadata>=4.0.0,<5.0.0;python_version<"3.8.0"',
		'typing_extensions>=3.7.4.2',
	],
	extras_require={
		'dev': [
			'mypy',
			'flake8<5.0',
			'flake8-annotations',
			'flake8-bandit',
			'flake8-bugbear',
			'flake8-commas',
			'flake8-comprehensions',
			'flake8-continuation',
			'flake8-datetimez',
			'flake8-docstrings',
			'flake8-import-order',
			'flake8-literal',
			'flake8-noqa',
			'flake8-polyfill',
			'flake8-postponed-annotations',
			'flake8-requirements',
			# 'flake8-smart-tabs',
			'flake8-tabs',
			'flake8-typechecking-import',
			'flake8-use-fstring',
			'pep8-naming',
		],
		'test': [],
	},
	classifiers=[
		'Framework :: Flake8',
		'Environment :: Console',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Topic :: Software Development :: Quality Assurance',
	],
	python_requires='>=3.7',
	entry_points={
		'flake8.extension': [
			f'{flake8_literal.quote_checker_prefix} = flake8_literal.quote_checker:QuoteChecker',
			f'{flake8_literal.raw_checker_prefix} = flake8_literal.raw_checker:RawChecker',
		],
	},
)
